# Generated by Django 2.0.2 on 2018-02-19 08:59

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("automated_logging", "0013_auto_20180218_1106"),
    ]

    operations = [
        migrations.AddField(
            model_name="request",
            name="status",
            field=models.PositiveSmallIntegerField(null=True),
        ),
        migrations.RemoveField(model_name="field", name="model"),
        migrations.AddField(
            model_name="field",
            name="model",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="dal_field",
                to="contenttypes.ContentType",
            ),
        ),
        migrations.DeleteModel(
            name="ModelStorage",
        ),
    ]



""" Test everything related to requests """

import json
from copy import deepcopy

from django.http import JsonResponse

from automated_logging.models import RequestEvent
from automated_logging.tests.base import BaseTestCase, USER_CREDENTIALS
from automated_logging.tests.helpers import random_string


class LoggedOutRequestsTestCase(BaseTestCase):
    def setUp(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        super().setUp()

        settings.AUTOMATED_LOGGING["request"]["exclude"]["applications"] = []
        conf.load.cache_clear()

        RequestEvent.objects.all().delete()

    @staticmethod
    def view(request):
        return JsonResponse({})

    def test_simple(self):
        self.bypass_request_restrictions()

        self.request("GET", self.view)

        events = RequestEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]

        self.assertEqual(event.user, None)


class LoggedInRequestsTestCase(BaseTestCase):
    def setUp(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        super().setUp()

        settings.AUTOMATED_LOGGING["request"]["exclude"]["applications"] = []
        conf.load.cache_clear()

        self.client.login(**USER_CREDENTIALS)

        RequestEvent.objects.all().delete()

    @staticmethod
    def view(request):
        return JsonResponse({})

    def test_simple(self):
        self.bypass_request_restrictions()

        self.request("GET", self.view)

        events = RequestEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]

        self.assertEqual(event.ip, "127.0.0.1")
        self.assertEqual(event.user, self.user)
        self.assertEqual(event.status, 200)
        self.assertEqual(event.method, "GET")
        self.assertEqual(event.uri, "/")

    def test_404(self):
        self.bypass_request_restrictions()

        self.client.get(f"/{random_string()}")

        events = RequestEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertEqual(event.status, 404)

    @staticmethod
    def exception(request):
        raise Exception

    def test_500(self):
        self.bypass_request_restrictions()

        try:
            self.request("GET", self.exception)
        except:
            pass

        events = RequestEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertGreaterEqual(event.status, 500)


class DataRecordingRequestsTestCase(BaseTestCase):
    def setUp(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        super().setUp()

        settings.AUTOMATED_LOGGING["request"]["exclude"]["applications"] = []
        settings.AUTOMATED_LOGGING["request"]["data"]["enabled"] = [
            "response",
            "request",
        ]
        conf.load.cache_clear()

        self.client.login(**USER_CREDENTIALS)

        RequestEvent.objects.all().delete()

    @staticmethod
    def view(request):
        return JsonResponse({"test": "example"})

    def test_payload(self):
        # TODO: preliminary until request/response parsing is implemented
        from django.conf import settings
        from automated_logging.settings import settings as conf

        self.bypass_request_restrictions()

        settings.AUTOMATED_LOGGING["request"]["data"]["enabled"] = [
            "request",
            "response",
        ]
        conf.load.cache_clear()

        self.request("GET", self.view, data=json.dumps({"X": "Y"}))

        events = RequestEvent.objects.all()
        self.assertEqual(events.count(), 1)

        response = json.dumps({"test": "example"})
        request = json.dumps({"X": "Y"})
        event = events[0]
        self.assertEqual(event.response.content.decode(), response)
        self.assertEqual(event.request.content.decode(), request)

    def test_exclusion_by_application(self):
        self.request("GET", self.view)
        self.assertEqual(RequestEvent.objects.count(), 0)



"""
Everything related to the admin interface of ModelMirror is located in here
"""

from django.contrib.admin import register
from django.shortcuts import redirect

from automated_logging.admin.base import ReadOnlyAdminMixin, ReadOnlyTabularInlineMixin
from automated_logging.models import ModelMirror, ModelField


class ModelFieldInline(ReadOnlyTabularInlineMixin):
    """list all recorded fields"""

    model = ModelField

    fields = ["name", "type"]

    verbose_name = "Recorded Field"
    verbose_name_plural = "Recorded Fields"


@register(ModelMirror)
class ModelMirrorAdmin(ReadOnlyAdminMixin):
    """admin page specification for ModelMirror"""

    def has_module_permission(self, request):
        """prevents this from showing up index.html"""
        return False

    def changelist_view(self, request, **kwargs):
        """instead of showing the changelist view redirect to the parent app_list"""
        return redirect("admin:app_list", self.model._meta.app_label)

    date_hierarchy = "updated_at"
    ordering = ("-updated_at",)

    fieldsets = (
        (
            "Information",
            {"fields": ("id", "application", "name")},
        ),
    )

    inlines = [ModelFieldInline]



# Generated by Django 3.0.7 on 2020-09-13 18:33

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):
    from django.conf import settings

    dependencies = [
        ("automated_logging", "0017_auto_20200819_1004"),
    ]

    if hasattr(settings, "AUTOMATED_LOGGING_DEV") and settings.AUTOMATED_LOGGING_DEV:
        operations = [
            migrations.CreateModel(
                name="DecoratorOverrideExclusionTest",
                fields=[
                    (
                        "id",
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ("created_at", models.DateTimeField(auto_now_add=True)),
                    ("updated_at", models.DateTimeField(auto_now=True)),
                    ("random", models.CharField(max_length=255, null=True)),
                    ("random2", models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name="FullClassBasedExclusionTest",
                fields=[
                    (
                        "id",
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ("created_at", models.DateTimeField(auto_now_add=True)),
                    ("updated_at", models.DateTimeField(auto_now=True)),
                    ("random", models.CharField(max_length=255, null=True)),
                    ("random2", models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name="FullDecoratorBasedExclusionTest",
                fields=[
                    (
                        "id",
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ("created_at", models.DateTimeField(auto_now_add=True)),
                    ("updated_at", models.DateTimeField(auto_now=True)),
                    ("random", models.CharField(max_length=255, null=True)),
                    ("random2", models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name="OrdinaryTest",
                fields=[
                    (
                        "id",
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ("created_at", models.DateTimeField(auto_now_add=True)),
                    ("updated_at", models.DateTimeField(auto_now=True)),
                    ("random", models.CharField(max_length=255, null=True)),
                    ("random2", models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name="PartialClassBasedExclusionTest",
                fields=[
                    (
                        "id",
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ("created_at", models.DateTimeField(auto_now_add=True)),
                    ("updated_at", models.DateTimeField(auto_now=True)),
                    ("random", models.CharField(max_length=255, null=True)),
                    ("random2", models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name="PartialDecoratorBasedExclusionTest",
                fields=[
                    (
                        "id",
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ("created_at", models.DateTimeField(auto_now_add=True)),
                    ("updated_at", models.DateTimeField(auto_now=True)),
                    ("random", models.CharField(max_length=255, null=True)),
                    ("random2", models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name="SpeedTest",
                fields=[
                    (
                        "id",
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ("created_at", models.DateTimeField(auto_now_add=True)),
                    ("updated_at", models.DateTimeField(auto_now=True)),
                    ("column0", models.CharField(max_length=15, null=True)),
                    ("column1", models.CharField(max_length=15, null=True)),
                    ("column2", models.CharField(max_length=15, null=True)),
                    ("column3", models.CharField(max_length=15, null=True)),
                    ("column4", models.CharField(max_length=15, null=True)),
                    ("column5", models.CharField(max_length=15, null=True)),
                    ("column6", models.CharField(max_length=15, null=True)),
                    ("column7", models.CharField(max_length=15, null=True)),
                    ("column8", models.CharField(max_length=15, null=True)),
                    ("column9", models.CharField(max_length=15, null=True)),
                    ("column10", models.CharField(max_length=15, null=True)),
                    ("column11", models.CharField(max_length=15, null=True)),
                    ("column12", models.CharField(max_length=15, null=True)),
                    ("column13", models.CharField(max_length=15, null=True)),
                    ("column14", models.CharField(max_length=15, null=True)),
                    ("column15", models.CharField(max_length=15, null=True)),
                    ("column16", models.CharField(max_length=15, null=True)),
                    ("column17", models.CharField(max_length=15, null=True)),
                    ("column18", models.CharField(max_length=15, null=True)),
                    ("column19", models.CharField(max_length=15, null=True)),
                    ("column20", models.CharField(max_length=15, null=True)),
                    ("column21", models.CharField(max_length=15, null=True)),
                    ("column22", models.CharField(max_length=15, null=True)),
                    ("column23", models.CharField(max_length=15, null=True)),
                    ("column24", models.CharField(max_length=15, null=True)),
                    ("column25", models.CharField(max_length=15, null=True)),
                    ("column26", models.CharField(max_length=15, null=True)),
                    ("column27", models.CharField(max_length=15, null=True)),
                    ("column28", models.CharField(max_length=15, null=True)),
                    ("column29", models.CharField(max_length=15, null=True)),
                    ("column30", models.CharField(max_length=15, null=True)),
                    ("column31", models.CharField(max_length=15, null=True)),
                    ("column32", models.CharField(max_length=15, null=True)),
                    ("column33", models.CharField(max_length=15, null=True)),
                    ("column34", models.CharField(max_length=15, null=True)),
                    ("column35", models.CharField(max_length=15, null=True)),
                    ("column36", models.CharField(max_length=15, null=True)),
                    ("column37", models.CharField(max_length=15, null=True)),
                    ("column38", models.CharField(max_length=15, null=True)),
                    ("column39", models.CharField(max_length=15, null=True)),
                    ("column40", models.CharField(max_length=15, null=True)),
                    ("column41", models.CharField(max_length=15, null=True)),
                    ("column42", models.CharField(max_length=15, null=True)),
                    ("column43", models.CharField(max_length=15, null=True)),
                    ("column44", models.CharField(max_length=15, null=True)),
                    ("column45", models.CharField(max_length=15, null=True)),
                    ("column46", models.CharField(max_length=15, null=True)),
                    ("column47", models.CharField(max_length=15, null=True)),
                    ("column48", models.CharField(max_length=15, null=True)),
                    ("column49", models.CharField(max_length=15, null=True)),
                    ("column50", models.CharField(max_length=15, null=True)),
                    ("column51", models.CharField(max_length=15, null=True)),
                    ("column52", models.CharField(max_length=15, null=True)),
                    ("column53", models.CharField(max_length=15, null=True)),
                    ("column54", models.CharField(max_length=15, null=True)),
                    ("column55", models.CharField(max_length=15, null=True)),
                    ("column56", models.CharField(max_length=15, null=True)),
                    ("column57", models.CharField(max_length=15, null=True)),
                    ("column58", models.CharField(max_length=15, null=True)),
                    ("column59", models.CharField(max_length=15, null=True)),
                    ("column60", models.CharField(max_length=15, null=True)),
                    ("column61", models.CharField(max_length=15, null=True)),
                    ("column62", models.CharField(max_length=15, null=True)),
                    ("column63", models.CharField(max_length=15, null=True)),
                    ("column64", models.CharField(max_length=15, null=True)),
                    ("column65", models.CharField(max_length=15, null=True)),
                    ("column66", models.CharField(max_length=15, null=True)),
                    ("column67", models.CharField(max_length=15, null=True)),
                    ("column68", models.CharField(max_length=15, null=True)),
                    ("column69", models.CharField(max_length=15, null=True)),
                    ("column70", models.CharField(max_length=15, null=True)),
                    ("column71", models.CharField(max_length=15, null=True)),
                    ("column72", models.CharField(max_length=15, null=True)),
                    ("column73", models.CharField(max_length=15, null=True)),
                    ("column74", models.CharField(max_length=15, null=True)),
                    ("column75", models.CharField(max_length=15, null=True)),
                    ("column76", models.CharField(max_length=15, null=True)),
                    ("column77", models.CharField(max_length=15, null=True)),
                    ("column78", models.CharField(max_length=15, null=True)),
                    ("column79", models.CharField(max_length=15, null=True)),
                    ("column80", models.CharField(max_length=15, null=True)),
                    ("column81", models.CharField(max_length=15, null=True)),
                    ("column82", models.CharField(max_length=15, null=True)),
                    ("column83", models.CharField(max_length=15, null=True)),
                    ("column84", models.CharField(max_length=15, null=True)),
                    ("column85", models.CharField(max_length=15, null=True)),
                    ("column86", models.CharField(max_length=15, null=True)),
                    ("column87", models.CharField(max_length=15, null=True)),
                    ("column88", models.CharField(max_length=15, null=True)),
                    ("column89", models.CharField(max_length=15, null=True)),
                    ("column90", models.CharField(max_length=15, null=True)),
                    ("column91", models.CharField(max_length=15, null=True)),
                    ("column92", models.CharField(max_length=15, null=True)),
                    ("column93", models.CharField(max_length=15, null=True)),
                    ("column94", models.CharField(max_length=15, null=True)),
                    ("column95", models.CharField(max_length=15, null=True)),
                    ("column96", models.CharField(max_length=15, null=True)),
                    ("column97", models.CharField(max_length=15, null=True)),
                    ("column98", models.CharField(max_length=15, null=True)),
                    ("column99", models.CharField(max_length=15, null=True)),
                ],
            ),
            migrations.CreateModel(
                name="OneToOneTest",
                fields=[
                    (
                        "id",
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ("created_at", models.DateTimeField(auto_now_add=True)),
                    ("updated_at", models.DateTimeField(auto_now=True)),
                    (
                        "relationship",
                        models.OneToOneField(
                            null=True,
                            on_delete=django.db.models.deletion.CASCADE,
                            to="automated_logging.OrdinaryTest",
                        ),
                    ),
                ],
            ),
            migrations.CreateModel(
                name="M2MTest",
                fields=[
                    (
                        "id",
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ("created_at", models.DateTimeField(auto_now_add=True)),
                    ("updated_at", models.DateTimeField(auto_now=True)),
                    (
                        "relationship",
                        models.ManyToManyField(to="automated_logging.OrdinaryTest"),
                    ),
                ],
            ),
            migrations.CreateModel(
                name="ForeignKeyTest",
                fields=[
                    (
                        "id",
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ("created_at", models.DateTimeField(auto_now_add=True)),
                    ("updated_at", models.DateTimeField(auto_now=True)),
                    (
                        "relationship",
                        models.ForeignKey(
                            null=True,
                            on_delete=django.db.models.deletion.CASCADE,
                            to="automated_logging.OrdinaryTest",
                        ),
                    ),
                ],
            ),
        ]

    else:
        operations = []



"""
Everything related to the admin interface of UnspecifiedEvent is located in here
"""

from django.contrib.admin import register, RelatedOnlyFieldListFilter

from automated_logging.admin.base import ReadOnlyAdminMixin, ReadOnlyTabularInlineMixin
from automated_logging.models import UnspecifiedEvent


@register(UnspecifiedEvent)
class UnspecifiedEventAdmin(ReadOnlyAdminMixin):
    """admin page specification for the UnspecifiedEvent"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_id(self, instance):
        """shorten the id to the first 8 digits"""
        return str(instance.id).split("-")[0]

    get_id.short_description = "UUID"

    list_display = ("get_id", "updated_at", "level", "message")

    date_hierarchy = "updated_at"
    ordering = ("-updated_at",)

    list_filter = ("updated_at",)

    fieldsets = (
        (
            "Information",
            {"fields": ("id", "updated_at", "application", "level", "message")},
        ),
        ("Location", {"fields": ("file", "line")}),
    )



import re
import typing
from collections import namedtuple
from datetime import timedelta
from typing import NamedTuple

from marshmallow import Schema, EXCLUDE, post_load
from marshmallow.fields import List, String, Nested, TimeDelta

from automated_logging.helpers.exceptions import NoMatchFound, CouldNotConvertError

Search = NamedTuple("Search", (("type", str), ("value", str)))
Search._serialize = lambda self: f"{self.type}:{self.value}"


class Set(List):
    """
    This is like a list, just compiles down to a set when serializing.
    """

    def _serialize(
        self, value, attr, obj, **kwargs
    ) -> typing.Optional[typing.Set[typing.Any]]:
        return set(super(Set, self)._serialize(value, attr, obj, **kwargs))

    def _deserialize(self, value, attr, data, **kwargs) -> typing.Set[typing.Any]:
        return set(super(Set, self)._deserialize(value, attr, data, **kwargs))


class LowerCaseString(String):
    """
    String that is always going to be serialized to a lowercase string,
    using `str.lower()`
    """

    def _deserialize(self, value, attr, data, **kwargs) -> str:
        output = super()._deserialize(value, attr, data, **kwargs)

        return output.lower()


class Duration(TimeDelta):
    """TimeDelta derivative, with more input methods"""

    def _convert(
        self, target: typing.Union[int, str, timedelta, None]
    ) -> typing.Optional[timedelta]:
        if target is None:
            return None

        if isinstance(target, timedelta):
            return target

        if isinstance(target, int) or isinstance(target, float):
            return timedelta(seconds=target)

        if isinstance(target, str):
            REGEX = (
                r"^P(?!$)(\d+Y)?(\d+M)?(\d+W)?(\d+D)?(T(?=\d)(\d+H)?(\d+M)?(\d+S)?)?$"
            )
            match = re.match(REGEX, target, re.IGNORECASE)
            if not match:
                raise self.make_error("invalid") from NoMatchFound

            components = list(match.groups())
            # remove leading T capture - isn't used, by removing the 5th capture group
            components.pop(4)

            adjusted = {"days": 0, "seconds": 0}
            conversion = [
                ["days", 365],  # year
                ["days", 30],  # month
                ["days", 7],  # week
                ["days", 1],  # day
                ["seconds", 3600],  # hour
                ["seconds", 60],  # minute
                ["seconds", 1],  # second
            ]

            for pointer in range(len(components)):
                if not components[pointer]:
                    continue
                rate = conversion[pointer]
                native = int(re.findall(r"(\d+)", components[pointer])[0])

                adjusted[rate[0]] += native * rate[1]

            return timedelta(**adjusted)

        raise self.make_error("invalid") from CouldNotConvertError

    def _deserialize(self, value, attr, data, **kwargs) -> typing.Optional[timedelta]:
        try:
            output = self._convert(value)
        except OverflowError as error:
            raise self.make_error("invalid") from error

        return output


class SearchString(String):
    """
    Used for:
    - ModelString
    - FieldString
    - ApplicationString
    - FileString

    SearchStrings are used for models, fields and applications.
    They can be either a glob (prefixed with either glob or gl),
    regex (prefixed with either regex or re)
    or plain (prefixed with plain or pl).

    All SearchStrings ignore the case of the raw string.

    format: <prefix>:<value>
    examples:
        - gl:app*       (glob matching)
        - glob:app*     (glob matching)
        - pl:app        (exact matching)
        - plain:app     (exact matching)
        - re:^app.*$    (regex matching)
        - regex:^app.*$ (regex matching)
        - :app*         (glob matching)
        - app           (glob matching)
    """

    def _deserialize(self, value, attr, data, **kwargs) -> Search:
        if isinstance(value, dict) and "type" in value and "value" in value:
            value = f'{value["type"]}:{value["value"]}'

        output = super()._deserialize(value, attr, data, **kwargs)

        match = re.match(r"^(\w*):(.*)$", output, re.IGNORECASE)
        if match:
            module = match.groups()[0].lower()
            match = match.groups()[1]

            if module.startswith("gl"):
                return Search("glob", match.lower())
            elif module.startswith("pl"):
                return Search("plain", match.lower())
            elif module.startswith("re"):
                # regex shouldn't be lowercase
                # we just ignore the case =
                return Search("regex", match)

            raise self.make_error("invalid") from NotImplementedError

        return Search("glob", output)


class MissingNested(Nested):
    """
    Modified marshmallow Nested, that is defaulting missing to loading an empty
    schema, to populate it with data.
    """

    def __init__(self, *args, **kwargs):
        if "missing" not in kwargs:
            kwargs["missing"] = lambda: args[0]().load({})

        super().__init__(*args, **kwargs)


class BaseSchema(Schema):
    """
    Modified marshmallow Schema, that is defaulting the unknown keyword to EXCLUDE,
    not RAISE (marshmallow default) and when loading converts the dict into a namedtuple.
    """

    def __init__(self, *args, **kwargs):
        if "unknown" not in kwargs:
            kwargs["unknown"] = EXCLUDE

        super().__init__(*args, **kwargs)

    @staticmethod
    def namedtuple_or(left: NamedTuple, right: NamedTuple):
        """
        __or__ implementation for the namedtuple
        """
        values = {}

        if not isinstance(left, tuple) or not isinstance(right, tuple):
            raise NotImplementedError

        for name in left._fields:
            field = getattr(left, name)
            values[name] = field

            if not hasattr(right, name):
                continue

            if isinstance(field, tuple) or isinstance(field, set):
                values[name] = field | getattr(right, name)

        return left._replace(**values)

    @staticmethod
    def namedtuple_factory(name, keys):
        """
        create the namedtuple from the name and keys to attach functions that are needed.

        Attaches:
            binary **or** operation to support globals propagation
        """
        Object = namedtuple(name, keys)
        Object.__or__ = BaseSchema.namedtuple_or
        return Object

    @post_load
    def make_namedtuple(self, data: typing.Dict, **kwargs):
        """
        converts the loaded data dict into a namedtuple

        :param data: loaded data
        :param kwargs: marshmallow kwargs
        :return: namedtuple
        """
        name = self.__class__.__name__.replace("Schema", "")

        Object = BaseSchema.namedtuple_factory(name, data.keys())
        return Object(**data)



# Generated by Django 2.1.1 on 2018-12-29 23:23

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):
    dependencies = [
        ("automated_logging", "0014_auto_20180219_0859"),
    ]

    operations = [
        migrations.DeleteModel(
            name="LDAP",
        ),
        migrations.AlterModelOptions(
            name="unspecified",
            options={
                "verbose_name": "Non DAL Message",
                "verbose_name_plural": "Non DAL Messages",
            },
        ),
        migrations.AlterField(
            model_name="application",
            name="id",
            field=models.UUIDField(
                db_index=True, default=uuid.uuid4, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="field",
            name="id",
            field=models.UUIDField(
                db_index=True, default=uuid.uuid4, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="model",
            name="id",
            field=models.UUIDField(
                db_index=True, default=uuid.uuid4, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="modelchangelog",
            name="id",
            field=models.UUIDField(
                db_index=True, default=uuid.uuid4, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="modelmodification",
            name="id",
            field=models.UUIDField(
                db_index=True, default=uuid.uuid4, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="modelobject",
            name="id",
            field=models.UUIDField(
                db_index=True, default=uuid.uuid4, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="request",
            name="id",
            field=models.UUIDField(
                db_index=True, default=uuid.uuid4, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="unspecified",
            name="id",
            field=models.UUIDField(
                db_index=True, default=uuid.uuid4, primary_key=True, serialize=False
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2018-02-16 10:04
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("automated_logging", "0005_auto_20180216_0941"),
    ]

    operations = [
        migrations.AlterField(
            model_name="model",
            name="application",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="atl_model_application",
                to="automated_logging.Application",
            ),
        ),
    ]



# Generated by Django 2.0.2 on 2018-02-16 15:45

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("automated_logging", "0010_auto_20180216_1430"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="model",
            options={"verbose_name": "Changelog", "verbose_name_plural": "Changelogs"},
        ),
        migrations.AlterModelOptions(
            name="modelchangelog",
            options={},
        ),
        migrations.AlterModelOptions(
            name="request",
            options={"verbose_name": "Request", "verbose_name_plural": "Requests"},
        ),
        migrations.AlterModelOptions(
            name="unspecified",
            options={
                "verbose_name": "Non DJL Message",
                "verbose_name_plural": "Non DJL Messages",
            },
        ),
    ]



# Generated by Django 2.0.2 on 2018-02-16 14:30

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):
    dependencies = [
        ("automated_logging", "0009_auto_20180216_1006"),
    ]

    operations = [
        migrations.CreateModel(
            name="Field",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ModelStorage",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="field",
            name="model",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="automated_logging.ModelStorage",
            ),
        ),
        migrations.AddField(
            model_name="modelobject",
            name="field",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="automated_logging.Field",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2018-02-15 15:40
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):
    dependencies = [
        ("automated_logging", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Application",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.RenameField(
            model_name="modelobject",
            old_name="application",
            new_name="type",
        ),
        migrations.RemoveField(
            model_name="modelobject",
            name="model",
        ),
        migrations.RemoveField(model_name="model", name="application"),
        migrations.AddField(
            model_name="model",
            name="application",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="atl_model_application",
                to="automated_logging.Application",
            ),
        ),
        migrations.RemoveField(model_name="request", name="application"),
        migrations.AddField(
            model_name="request",
            name="application",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="automated_logging.Application",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2018-02-16 09:35
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("automated_logging", "0003_auto_20180216_0900"),
    ]

    operations = [
        migrations.AlterField(
            model_name="modelobject",
            name="type",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="atl_modelobject_application",
                to="contenttypes.ContentType",
            ),
        ),
    ]



"""
This module specifically handlers "many to many" changes, those are
a bit more complicated as we need to detect the changes
on a per field basis.

This finds the changes and redirects them to the handler,
without doing any changes to the database.
"""

import logging
from typing import Optional

from django.db.models import Manager
from django.db.models.fields.related import ManyToManyField
from django.db.models.signals import m2m_changed
from django.dispatch import receiver

from automated_logging.helpers import (
    Operation,
    get_or_create_model_event,
    get_or_create_meta,
)
from automated_logging.models import (
    ModelRelationshipModification,
    ModelEntry,
    ModelMirror,
    Application,
    ModelField,
)
from automated_logging.settings import settings
from automated_logging.signals import lazy_model_exclusion

logger = logging.getLogger(__name__)


def find_m2m_rel(sender, model) -> Optional[ManyToManyField]:
    """
    This finds the "many to many" relationship that is used by the sender.
    """
    for field in model._meta.get_fields():
        if isinstance(field, ManyToManyField) and field.remote_field.through == sender:
            return field

    return None


def post_processor(sender, instance, model, operation, targets):
    """
    if the change is in reverse or not, the processing of the changes is still
    the same, so we have this method to take care of constructing the changes
    :param sender:
    :param instance:
    :param model:
    :param operation:
    :param targets:
    :return:
    """
    relationships = []

    m2m_rel = find_m2m_rel(sender, model)
    if not m2m_rel:
        logger.warning(f"[DAL] save[m2m] could not find ManyToManyField for {instance}")
        return

    field = ModelField()
    field.name = m2m_rel.name
    field.mirror = ModelMirror(
        name=model.__name__, application=Application(name=instance._meta.app_label)
    )
    field.type = m2m_rel.__class__.__name__

    # there is the possibility that a pre_clear occurred, if that is the case
    # extend the targets and pop the list of affected instances from the attached
    # field
    get_or_create_meta(instance)
    if (
        hasattr(instance._meta.dal, "m2m_pre_clear")
        and field.name in instance._meta.dal.m2m_pre_clear
        and operation == Operation.DELETE
    ):
        cleared = instance._meta.dal.m2m_pre_clear[field.name]
        targets.extend(cleared)
        instance._meta.dal.m2m_pre_clear.pop(field.name)

    for target in targets:
        relationship = ModelRelationshipModification()
        relationship.operation = operation
        relationship.field = field
        mirror = ModelMirror()
        mirror.name = target.__class__.__name__
        mirror.application = Application(name=target._meta.app_label)
        relationship.entry = ModelEntry(
            mirror=mirror, value=repr(target), primary_key=target.pk
        )
        relationships.append(relationship)

    if len(relationships) == 0:
        # there was no actual change, so we're not propagating the event
        return

    event, _ = get_or_create_model_event(instance, operation)

    user = None
    logger.log(
        settings.model.loglevel,
        f'{user or "Anonymous"} modified field '
        f"{field.name} | Model: "
        f"{field.mirror.application}.{field.mirror} "
        f'| Modifications: {", ".join([r.short() for r in relationships])}',
        extra={
            "action": "model[m2m]",
            "data": {"instance": instance, "sender": sender},
            "relationships": relationships,
            "event": event,
        },
    )


def pre_clear_processor(sender, instance, pks, model, reverse, operation) -> None:
    """
    pre_clear needs a specific processor as we attach the changes that are about
    to happen to the instance first, and then use them in post_delete/post_clear

    if reverse = False then every element gets removed from the relationship field,
    but if reverse = True then instance should be removed from every target.

    Note: it seems that pre_clear is not getting fired for reverse.

    :return: None
    """
    if reverse:
        return

    get_or_create_meta(instance)

    rel = find_m2m_rel(sender, instance.__class__)
    if "m2m_pre_clear" not in instance._meta.dal:
        instance._meta.dal.m2m_pre_clear = {}

    cleared = getattr(instance, rel.name, [])
    if isinstance(cleared, Manager):
        cleared = list(cleared.all())
    instance._meta.dal.m2m_pre_clear = {rel.name: cleared}


@receiver(m2m_changed, weak=False)
def m2m_changed_signal(
    sender, instance, action, reverse, model, pk_set, using, **kwargs
) -> None:
    """
    Django sends this signal when many-to-many relationships change.

    One of the more complex signals, due to the fact that change can be reversed
    we need to either process
    instance field changes of pk_set (reverse=False) or
    pk_set field changes of instance. (reverse=True)

    The changes will always get applied in the model where the field in defined.

    # TODO: post_remove also gets triggered when there is nothing actually getting removed
    :return: None
    """
    if action not in ["post_add", "post_remove", "pre_clear", "post_clear"]:
        return

    if action == "pre_clear":
        operation = Operation.DELETE

        return pre_clear_processor(
            sender,
            instance,
            list(pk_set) if pk_set else None,
            model,
            reverse,
            operation,
        )
    elif action == "post_add":
        operation = Operation.CREATE
    elif action == "post_clear":
        operation = Operation.DELETE
    else:
        operation = Operation.DELETE

    targets = model.objects.filter(pk__in=list(pk_set)) if pk_set else []
    if reverse:
        for target in [
            t for t in targets if not lazy_model_exclusion(t, operation, t.__class__)
        ]:
            post_processor(sender, target, target.__class__, operation, [instance])
    else:
        if lazy_model_exclusion(instance, operation, instance.__class__):
            return

        post_processor(sender, instance, instance.__class__, operation, targets)



""" Test the save functionality """

import datetime

from django.http import JsonResponse

from automated_logging.helpers import Operation
from automated_logging.models import ModelEvent
from automated_logging.tests.base import BaseTestCase, USER_CREDENTIALS
from automated_logging.tests.helpers import random_string
from automated_logging.tests.models import OrdinaryTest


class LoggedOutSaveModificationsTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()

        # delete all previous model events
        ModelEvent.objects.all().delete()

    def test_create_simple_value(self):
        """
        test if creation results in the correct fields
        :return:
        """
        self.bypass_request_restrictions()
        value = random_string()

        instance = OrdinaryTest()
        instance.random = value
        instance.save()

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]

        self.assertEqual(event.operation, int(Operation.CREATE))
        self.assertEqual(event.user, None)

        self.assertEqual(event.entry.primary_key, str(instance.pk))
        self.assertEqual(event.entry.value, repr(instance))

        self.assertEqual(event.entry.mirror.name, "OrdinaryTest")
        self.assertEqual(event.entry.mirror.application.name, "automated_logging")

        modifications = event.modifications.all()
        # pk and random added and modified
        self.assertEqual(modifications.count(), 2)
        self.assertEqual({m.field.name for m in modifications}, {"random", "id"})

        modification = [m for m in modifications if m.field.name == "random"][0]
        self.assertEqual(modification.operation, int(Operation.CREATE))
        self.assertEqual(modification.previous, None)
        self.assertEqual(modification.current, value)
        self.assertEqual(modification.event, event)

        self.assertEqual(modification.field.name, "random")
        self.assertEqual(modification.field.type, "CharField")

        relationships = event.relationships.all()
        self.assertEqual(relationships.count(), 0)

    def test_modify(self):
        """
        test if modification results
        in proper delete and create field operations
        :return:
        """
        previous, current = random_string(10), random_string(10)

        # create instance
        instance = OrdinaryTest()
        instance.random = previous
        instance.save()

        # delete all stuff related to the instance events to have a clean slate
        ModelEvent.objects.all().delete()

        instance.random = current
        instance.save()

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]

        self.assertEqual(event.user, None)
        self.assertEqual(event.operation, int(Operation.MODIFY))

        modifications = event.modifications.all()
        self.assertEqual(modifications.count(), 1)

        modification = modifications[0]
        self.assertEqual(modification.operation, int(Operation.MODIFY))
        self.assertEqual(modification.previous, previous)
        self.assertEqual(modification.current, current)

        relationships = event.relationships.all()
        self.assertEqual(relationships.count(), 0)

    def test_honor_save(self):
        """
        test if saving honors the only attribute

        :return:
        """
        previous1, random1, random2 = (
            random_string(10),
            random_string(10),
            random_string(10),
        )

        # create instance
        instance = OrdinaryTest()
        instance.random = previous1
        instance.save()

        ModelEvent.objects.all().delete()

        instance.random = random1
        instance.random2 = random2
        instance.save(update_fields=["random2"])

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        modifications = event.modifications.all()
        self.assertEqual(modifications.count(), 1)

        modification = modifications[0]
        self.assertEqual(modification.operation, int(Operation.CREATE))
        self.assertEqual(modification.field.name, "random2")

    def test_delete(self):
        """
        test if deletion is working correctly and records all the changes
        :return:
        """
        value = random_string(10)

        # create instance
        instance = OrdinaryTest()
        instance.random = value
        instance.save()

        pk = instance.pk
        re = repr(instance)

        ModelEvent.objects.all().delete()

        instance.delete()

        # DUP of save
        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertEqual(event.operation, int(Operation.DELETE))
        self.assertEqual(event.user, None)

        self.assertEqual(event.entry.primary_key, str(pk))
        self.assertEqual(event.entry.value, re)

        # DELETE does currently not record deleted values
        modifications = event.modifications.all()
        self.assertEqual(modifications.count(), 0)

    def test_reproducibility(self):
        """
        test if all the changes where done
        correctly so that you can properly derive the
        current state from all the accumulated changes

        TODO
        """
        pass

    def test_performance(self):
        """
        test if setting the performance parameter works correctly

        :return:
        """
        from django.conf import settings
        from automated_logging.settings import settings as conf

        self.bypass_request_restrictions()

        settings.AUTOMATED_LOGGING["model"]["performance"] = True
        conf.load.cache_clear()

        ModelEvent.objects.all().delete()
        instance = OrdinaryTest()
        instance.random = random_string(10)
        checkpoint = datetime.datetime.now()
        instance.save()
        checkpoint = datetime.datetime.now() - checkpoint

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertIsNotNone(event.performance)
        self.assertLess(event.performance.total_seconds(), checkpoint.total_seconds())

    def test_snapshot(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        self.bypass_request_restrictions()

        settings.AUTOMATED_LOGGING["model"]["snapshot"] = True
        conf.load.cache_clear()

        instance = OrdinaryTest(random=random_string())
        instance.save()

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertIsNotNone(event.snapshot)
        self.assertEqual(instance, event.snapshot)


class LoggedInSaveModificationsTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.client.login(**USER_CREDENTIALS)

        self.clear()

    @staticmethod
    def view(request):
        value = random_string()

        instance = OrdinaryTest()
        instance.random = value
        instance.save()

        return JsonResponse({})

    def test_user(self):
        """Test if DAL recognizes the user through the middleware"""
        self.bypass_request_restrictions()

        response = self.request("GET", self.view)
        self.assertEqual(response.content, b"{}")

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertEqual(event.user, self.user)


# TODO: test snapshot



"""
Everything related to the admin interface of ModelEntry is located in here
"""

from django.contrib.admin import register
from django.shortcuts import redirect

from automated_logging.admin.model_event import ModelEventAdmin
from automated_logging.admin.base import ReadOnlyAdminMixin, ReadOnlyTabularInlineMixin
from automated_logging.models import ModelEntry, ModelEvent


class ModelEventInline(ReadOnlyTabularInlineMixin):
    """inline for all attached events"""

    model = ModelEvent

    def __init__(self, *args, **kwargs):
        super(ModelEventInline, self).__init__(*args, **kwargs)

        self.readonly_fields = [*self.readonly_fields, "get_uuid", "get_modifications"]

    def get_uuid(self, instance):
        """make the uuid small"""
        return self.model_admin_url(instance, str(instance.id).split("-")[0])

    get_uuid.short_description = "UUID"

    def get_modifications(self, instance):
        """ModelEventAdmin already implements this functions, we just refer to it"""
        return ModelEventAdmin.get_modifications(self, instance)

    get_modifications.short_description = "Modifications"

    fields = ("get_uuid", "updated_at", "user", "get_modifications")

    ordering = ("-updated_at",)

    verbose_name = "Event"
    verbose_name_plural = "Events"


@register(ModelEntry)
class ModelEntryAdmin(ReadOnlyAdminMixin):
    """admin page specification for ModelEntry"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.readonly_fields = [*self.readonly_fields, "get_model", "get_application"]

    def changelist_view(self, request, **kwargs):
        """instead of showing the changelist view redirect to the parent app_list"""
        return redirect("admin:app_list", self.model._meta.app_label)

    def has_module_permission(self, request):
        """remove model entries from the index.html list"""
        return False

    def get_model(self, instance):
        """get the model mirror"""
        return self.model_admin_url(instance.mirror)

    get_model.short_description = "Model"

    def get_application(self, instance):
        """get the application"""
        return instance.mirror.application

    get_application.short_description = "Application"

    fieldsets = (
        (
            "Information",
            {"fields": ("id", "get_model", "get_application", "primary_key", "value")},
        ),
    )

    inlines = [ModelEventInline]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.2 on 2018-02-15 14:22
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="LDAP",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("action", models.TextField()),
                ("succeeded", models.NullBooleanField()),
                ("errorMessage", models.TextField(blank=True, null=True)),
                ("basedn", models.TextField(blank=True, null=True)),
                ("entry", models.TextField(blank=True, null=True)),
                ("objectClass", models.TextField(blank=True, null=True)),
                ("cn", models.TextField(blank=True, null=True)),
                ("existing_members", models.TextField(blank=True, null=True)),
                ("data_members", models.TextField(blank=True, null=True)),
                ("diff_members", models.TextField(blank=True, null=True)),
            ],
            options={
                "verbose_name": "LDAP event log entry",
                "verbose_name_plural": "LDAP event log entries",
            },
        ),
        migrations.CreateModel(
            name="Model",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("message", models.TextField(null=True)),
                (
                    "action",
                    models.PositiveSmallIntegerField(
                        choices=[(0, "n/a"), (1, "add"), (2, "change"), (3, "delete")],
                        default=0,
                    ),
                ),
                (
                    "application",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="atl_model_application",
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
            options={
                "verbose_name": "model log entry",
                "verbose_name_plural": "model log entries",
            },
        ),
        migrations.CreateModel(
            name="ModelChangelog",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "Model entry change",
                "verbose_name_plural": "Model entry changes",
            },
        ),
        migrations.CreateModel(
            name="ModelModification",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ModelObject",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("value", models.CharField(max_length=255)),
                ("model", models.CharField(max_length=255)),
                (
                    "application",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="atl_modelobject_application",
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Request",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("url", models.URLField()),
                (
                    "application",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "request event entry",
                "verbose_name_plural": "request event entries",
            },
        ),
        migrations.CreateModel(
            name="Unspecified",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("message", models.TextField(null=True)),
                ("level", models.PositiveSmallIntegerField(default=20)),
                ("file", models.CharField(max_length=255, null=True)),
                ("line", models.PositiveIntegerField(null=True)),
            ],
            options={
                "verbose_name": " logging entry (Errors, Warnings, Info)",
                "verbose_name_plural": " logging entries (Errors, Warnings, Info)",
            },
        ),
        migrations.AddField(
            model_name="modelmodification",
            name="currently",
            field=models.ManyToManyField(
                related_name="changelog_current", to="automated_logging.ModelObject"
            ),
        ),
        migrations.AddField(
            model_name="modelmodification",
            name="previously",
            field=models.ManyToManyField(
                related_name="changelog_previous", to="automated_logging.ModelObject"
            ),
        ),
        migrations.AddField(
            model_name="modelchangelog",
            name="information",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                to="automated_logging.ModelObject",
            ),
        ),
        migrations.AddField(
            model_name="modelchangelog",
            name="inserted",
            field=models.ManyToManyField(
                related_name="changelog_inserted", to="automated_logging.ModelObject"
            ),
        ),
        migrations.AddField(
            model_name="modelchangelog",
            name="modification",
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="automated_logging.ModelModification",
            ),
        ),
        migrations.AddField(
            model_name="modelchangelog",
            name="removed",
            field=models.ManyToManyField(
                related_name="changelog_removed", to="automated_logging.ModelObject"
            ),
        ),
        migrations.AddField(
            model_name="model",
            name="information",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="atl_model_information",
                to="automated_logging.ModelObject",
            ),
        ),
        migrations.AddField(
            model_name="model",
            name="modification",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="automated_logging.ModelChangelog",
            ),
        ),
        migrations.AddField(
            model_name="model",
            name="user",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



import uuid
from django.db.models import (
    CASCADE,
    CharField,
    DateTimeField,
    ForeignKey,
    ManyToManyField,
    Model,
    OneToOneField,
    UUIDField,
)

from automated_logging.decorators import exclude_model, include_model


class TestBase(Model):
    """Base for all the test models"""

    id = UUIDField(default=uuid.uuid4, primary_key=True)

    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    class Meta:
        abstract = True
        app_label = "automated_logging"


class OrdinaryBaseTest(TestBase):
    """Ordinary base test. Has a random char field."""

    random = CharField(max_length=255, null=True)
    random2 = CharField(max_length=255, null=True)

    class Meta:
        abstract = True
        app_label = "automated_logging"


class OrdinaryTest(OrdinaryBaseTest):
    """Ordinary test. Has a random char field."""

    class Meta:
        app_label = "automated_logging"


class M2MTest(TestBase):
    """Used to test the Many-To-Many Relationship functionality of DAL"""

    relationship = ManyToManyField(OrdinaryTest)

    class Meta:
        app_label = "automated_logging"


class ForeignKeyTest(TestBase):
    """Used to test ForeignKey functionality of DAL."""

    relationship = ForeignKey(OrdinaryTest, on_delete=CASCADE, null=True)

    class Meta:
        app_label = "automated_logging"


class OneToOneTest(TestBase):
    """Used to test the One-To-One Relationship functionality of DAL."""

    relationship = OneToOneField(OrdinaryTest, on_delete=CASCADE, null=True)

    class Meta:
        app_label = "automated_logging"


class SpeedTest(TestBase):
    """Used to test the speed of DAL"""

    for idx in range(100):
        exec(f"column{idx} = CharField(max_length=15, null=True)")

    class Meta:
        app_label = "automated_logging"


class FullClassBasedExclusionTest(OrdinaryBaseTest):
    """Used to test the full model exclusion via meta class"""

    class Meta:
        app_label = "automated_logging"

    class LoggingIgnore:
        complete = True


class PartialClassBasedExclusionTest(OrdinaryBaseTest):
    """Used to test partial ignore via fields"""

    class Meta:
        app_label = "automated_logging"

    class LoggingIgnore:
        fields = ["random"]
        operations = ["delete"]


@exclude_model
class FullDecoratorBasedExclusionTest(OrdinaryBaseTest):
    """Used to test full decorator exclusion"""

    class Meta:
        app_label = "automated_logging"


@exclude_model(operations=["delete"], fields=["random"])
class PartialDecoratorBasedExclusionTest(OrdinaryBaseTest):
    """Used to test partial decorator exclusion"""

    class Meta:
        app_label = "automated_logging"


@include_model
class DecoratorOverrideExclusionTest(OrdinaryBaseTest):
    """
    Used to check if include_model
    has precedence over class based configuration
    """

    class Meta:
        app_label = "automated_logging"

    class LoggingIgnore:
        complete = True



"""
File handles the processing and redirection of all request related
signals
"""

import logging
import urllib.parse

from django.core.handlers.wsgi import WSGIRequest
from django.dispatch import receiver
from django.core.signals import got_request_exception, request_finished
from django.http import Http404, JsonResponse
from django.urls import resolve

from automated_logging.middleware import AutomatedLoggingMiddleware
from automated_logging.models import RequestEvent, Application, RequestContext
from automated_logging.settings import settings
from automated_logging.signals import request_exclusion

# TODO: should django-ipware be optional?
try:
    from ipware import get_client_ip
except ImportError:
    get_client_ip = None


logger = logging.getLogger(__name__)


@receiver(request_finished, weak=False)
def request_finished_signal(sender, **kwargs) -> None:
    """
    This signal gets the environment from the local thread and
    sends a logging message, that message will be processed by the
    handler later on.

    This is a simple redirection.

    :return: -
    """
    level = settings.request.loglevel
    environ = AutomatedLoggingMiddleware.get_current_environ()

    if not environ:
        if settings.request.log_request_was_not_recorded:
            logger.info(
                "Environment for request couldn't be determined. "
                "Request was not recorded."
            )
        return

    request = RequestEvent()

    request.user = AutomatedLoggingMiddleware.get_current_user(environ)
    request.uri = environ.request.get_full_path()

    if not settings.request.data.query:
        request.uri = urllib.parse.urlparse(request.uri).path

    if "request" in settings.request.data.enabled:
        request_context = RequestContext()
        request_context.content = environ.request.body
        request_context.type = environ.request.content_type

        request.request = request_context

    if "response" in settings.request.data.enabled:
        response_context = RequestContext()
        response_context.content = environ.response.content
        response_context.type = environ.response["Content-Type"]

        request.response = response_context

    # TODO: context parsing, masking and removal
    if get_client_ip and settings.request.ip:
        request.ip, _ = get_client_ip(environ.request)

    request.status = environ.response.status_code if environ.response else None
    request.method = environ.request.method.upper()
    request.context_type = environ.request.content_type

    try:
        function = resolve(environ.request.path).func
    except Http404:
        function = None

    request.application = Application(name=None)
    if function:
        application = function.__module__.split(".")[0]
        request.application = Application(name=application)

    if request_exclusion(request, function):
        return

    logger_ip = f" from {request.ip}" if get_client_ip and settings.request.ip else ""
    logger.log(
        level,
        f"[{request.method}] [{request.status}] "
        f'{getattr(request, "user", None) or "Anonymous"} '
        f"at {request.uri}{logger_ip}",
        extra={"action": "request", "event": request},
    )


@receiver(got_request_exception, weak=False)
def request_exception(sender, request, **kwargs):
    """
    Exception logging for requests, via the django signal.

    The signal can also return a WSGIRequest exception, which does not
    have all fields that are needed.

    :return: -
    """

    status = int(request.status_code) if hasattr(request, "status_code") else None
    method = request.method if hasattr(request, "method") else None
    reason = request.reason_phrase if hasattr(request, "reason_phrase") else None
    level = logging.CRITICAL if status and status <= 500 else logging.WARNING

    is_wsgi = isinstance(request, WSGIRequest)

    logger.log(
        level,
        f'[{method or "UNK"}] [{status or "UNK"}] '
        f'{is_wsgi and "[WSGIResponse] "}'
        f'Exception: {reason or "UNKNOWN"}',
    )


@receiver(request_finished, weak=False)
def thread_cleanup(sender, **kwargs):
    """
    This signal just calls the thread cleanup function to make sure,
    that the custom thread object is always clean for the next request.
    This needs to be always the last function registered by the receiver!

    :return: -
    """
    AutomatedLoggingMiddleware.cleanup()



# test max_age (rework?)
# test save
# test module removal
import logging
import logging.config
from datetime import timedelta
import time

from django.http import JsonResponse
from marshmallow import ValidationError

from automated_logging.helpers.exceptions import CouldNotConvertError
from automated_logging.models import ModelEvent, RequestEvent, UnspecifiedEvent
from automated_logging.tests.models import OrdinaryTest
from automated_logging.tests.base import BaseTestCase


class TestDatabaseHandlerTestCase(BaseTestCase):
    @staticmethod
    def view(request):
        return JsonResponse({})

    def test_max_age(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        duration = timedelta(seconds=1)
        logger = logging.getLogger(__name__)

        settings.AUTOMATED_LOGGING["model"]["max_age"] = duration
        settings.AUTOMATED_LOGGING["request"]["max_age"] = duration
        settings.AUTOMATED_LOGGING["unspecified"]["max_age"] = duration

        conf.load.cache_clear()

        self.clear()
        self.bypass_request_restrictions()

        OrdinaryTest().save()
        self.request("GET", self.view)
        logger.info("I have the high ground Anakin!")

        self.assertEqual(ModelEvent.objects.count(), 1)
        self.assertEqual(RequestEvent.objects.count(), 1)
        self.assertEqual(UnspecifiedEvent.objects.count(), 1)

        time.sleep(2)

        logger.info("A surprise, to be sure, but a welcome one.")

        self.assertEqual(ModelEvent.objects.count(), 0)
        self.assertEqual(RequestEvent.objects.count(), 0)
        self.assertEqual(UnspecifiedEvent.objects.count(), 1)

    def test_max_age_input_methods(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        logger = logging.getLogger(__name__)

        settings.AUTOMATED_LOGGING["unspecified"]["max_age"] = timedelta(seconds=1)
        conf.load.cache_clear()
        self.clear()

        logger.info("I will do what I must.")
        time.sleep(1)
        logger.info("Hello There.")
        self.assertEqual(UnspecifiedEvent.objects.count(), 1)

        settings.AUTOMATED_LOGGING["unspecified"]["max_age"] = 1
        conf.load.cache_clear()
        self.clear()

        logger.info("A yes, the negotiator.")
        time.sleep(1)
        logger.info("Your tactics confuse and frighten me, sir.")
        self.assertEqual(UnspecifiedEvent.objects.count(), 1)

        settings.AUTOMATED_LOGGING["unspecified"]["max_age"] = "PT1S"
        conf.load.cache_clear()
        self.clear()

        logger.info("Don't make me kill you.")
        time.sleep(1)
        logger.info("An old friend from the dead.")
        self.assertEqual(UnspecifiedEvent.objects.count(), 1)

    def test_batching(self):
        from django.conf import settings

        logger = logging.getLogger(__name__)

        config = settings.LOGGING

        config["handlers"]["db"]["batch"] = 10
        logging.config.dictConfig(config)

        self.clear()
        for _ in range(9):
            logger.info("It's a trick. Send no reply")

        self.assertEqual(UnspecifiedEvent.objects.count(), 0)
        logger.info("I can't see a thing. My cockpit's fogging")
        self.assertEqual(UnspecifiedEvent.objects.count(), 10)

        config["handlers"]["db"]["batch"] = 1
        logging.config.dictConfig(config)



""" custom exceptions """


class NoMatchFound(Exception):
    """error that indicates that no match has been found for a regex"""

    pass


class CouldNotConvertError(Exception):
    """error that is thrown when no conversion could be done"""

    pass



from django.contrib.admin.options import BaseModelAdmin, ModelAdmin, TabularInline
from django.contrib.admin.templatetags.admin_urls import admin_urlname
from django.shortcuts import resolve_url
from django.utils.html import format_html
from django.utils.safestring import SafeText

from automated_logging.models import BaseModel


class MixinBase(BaseModelAdmin):
    """
    TabularInline and ModelAdmin readonly mixin have both the same methods and
    return the same, because of that fact we have a mixin base
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.readonly_fields = [f.name for f in self.model._meta.get_fields()]

    def get_actions(self, request):
        """get_actions from ModelAdmin, but remove all write operations."""
        actions = super().get_actions(request)
        actions.pop("delete_selected", None)

        return actions

    def has_add_permission(self, request, instance=None):
        """no-one should have the ability to add something => r/o"""
        return False

    def has_delete_permission(self, request, instance=None):
        """no-one should have the ability to delete something => r/o"""
        return False

    def has_change_permission(self, request, instance=None):
        """no-one should have the ability to edit something => r/o"""
        return False

    def save_model(self, request, instance, form, change):
        """disable saving by doing nothing"""
        pass

    def delete_model(self, request, instance):
        """disable deleting by doing nothing"""
        pass

    def save_related(self, request, form, formsets, change):
        """we don't need to save related, because save_model does nothing"""
        pass

    # helpers
    def model_admin_url(self, instance: BaseModel, name: str = None) -> str:
        """Helper to return a URL to another object"""
        url = resolve_url(
            admin_urlname(instance._meta, SafeText("change")), instance.pk
        )
        return format_html('<a href="{}">{}</a>', url, name or str(instance))


class ReadOnlyAdminMixin(MixinBase, ModelAdmin):
    """Disables all editing capabilities for the model admin"""

    change_form_template = "dal/admin/view.html"


class ReadOnlyTabularInlineMixin(MixinBase, TabularInline):
    """Disables all editing capabilities for inline"""

    model = None



# Generated by Django 3.0.7 on 2020-08-19 10:04

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("automated_logging", "0016_auto_20200803_1917"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="application",
            options={
                "verbose_name": "Application",
                "verbose_name_plural": "Applications",
            },
        ),
    ]



""" This is just a file that imports all admin interfaces defined in this package """

from automated_logging.admin.model_entry import *
from automated_logging.admin.model_event import *
from automated_logging.admin.model_mirror import *

from automated_logging.admin.request_event import *

from automated_logging.admin.unspecified_event import *



"""
Everything related to the admin interface of RequestEvent is located in here
"""

from django.contrib.admin import register, RelatedOnlyFieldListFilter

from automated_logging.admin.base import ReadOnlyAdminMixin, ReadOnlyTabularInlineMixin
from automated_logging.models import RequestEvent


@register(RequestEvent)
class RequestEventAdmin(ReadOnlyAdminMixin):
    """admin page specification for the RequestEvent"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.readonly_fields = [*self.readonly_fields, "get_uri", "get_user"]

    def get_id(self, instance):
        """shorten the id to the first 8 digits"""
        return str(instance.id).split("-")[0]

    get_id.short_description = "UUID"

    def get_uri(self, instance):
        """get the uri. just a redirect to set the short description."""
        return instance.uri

    get_uri.short_description = "URI"

    def get_user(self, instance):
        """get the user with a URL"""
        return self.model_admin_url(instance.user)

    get_user.short_description = "User"

    list_display = ("get_id", "updated_at", "user", "method", "status", "uri")

    date_hierarchy = "updated_at"
    ordering = ("-updated_at",)

    list_filter = ("updated_at", ("user", RelatedOnlyFieldListFilter))

    fieldsets = (
        (
            "Information",
            {
                "fields": (
                    "id",
                    "get_user",
                    "updated_at",
                    "application",
                )
            },
        ),
        ("HTTP", {"fields": ("method", "status", "get_uri")}),
    )
    # TODO: Context



from django.apps import AppConfig
from .settings import settings


class AutomatedloggingConfig(AppConfig):
    name = "automated_logging"
    verbose_name = "Django Automated Logging (DAL)"

    def ready(self):
        if "request" in settings.modules:
            from .signals import request
        if "model" in settings.modules:
            from .signals import save
            from .signals import m2m

        from .handlers import DatabaseHandler



"""
Everything related to the admin interface of ModelEvent is located in here
"""

from django.contrib.admin import register, RelatedOnlyFieldListFilter
from django.utils.html import format_html

from automated_logging.admin.base import ReadOnlyTabularInlineMixin, ReadOnlyAdminMixin
from automated_logging.helpers import Operation
from automated_logging.models import (
    ModelValueModification,
    ModelRelationshipModification,
    ModelEvent,
)


class ModelValueModificationInline(ReadOnlyTabularInlineMixin):
    """inline for all modifications"""

    model = ModelValueModification

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.readonly_fields = [*self.readonly_fields, "get_uuid", "get_field"]

    def get_uuid(self, instance):
        """make the uuid small"""
        return str(instance.id).split("-")[0]

    get_uuid.short_description = "UUID"

    def get_field(self, instance):
        """show the field name"""
        return instance.field.name

    get_field.short_description = "Field"

    fields = ("get_uuid", "operation", "get_field", "previous", "current")
    can_delete = False

    verbose_name = "Modification"
    verbose_name_plural = "Modifications"


class ModelRelationshipModificationInline(ReadOnlyTabularInlineMixin):
    """inline for all relationship modifications"""

    model = ModelRelationshipModification

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.readonly_fields = [*self.readonly_fields, "get_uuid", "get_field"]

    def get_uuid(self, instance):
        """make the uuid small"""
        return str(instance.id).split("-")[0]

    get_uuid.short_description = "UUID"

    def get_field(self, instance):
        """show the field name"""
        return instance.field.name

    get_field.short_description = "Field"

    fields = ("get_uuid", "operation", "get_field", "entry")
    can_delete = False

    verbose_name = "Relationship"
    verbose_name_plural = "Relationships"


@register(ModelEvent)
class ModelEventAdmin(ReadOnlyAdminMixin):
    """admin page specification for ModelEvent"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.readonly_fields = [
            *self.readonly_fields,
            "get_application",
            "get_user",
            "get_model_link",
        ]

    def get_modifications(self, instance):
        """
        Modifications in short form, are colored for better readability.

        Colors taken from
        https://github.com/django/django/tree/master/django/contrib/admin/static/admin/img
        """
        colors = {
            Operation.CREATE: "#70bf2b",
            Operation.MODIFY: "#efb80b",
            Operation.DELETE: "#dd4646",
        }
        return format_html(
            ", ".join(
                [
                    *[
                        f'<span style="color: {colors[Operation(m.operation)]};">'
                        f"{m.short()}"
                        f"</span>"
                        for m in instance.modifications.all()
                    ],
                    *[
                        f'<span style="color: {colors[Operation(r.operation)]};">'
                        f"{r.medium()[0]}"
                        f"</span>[{r.medium()[1]}]"
                        for r in instance.relationships.all()
                    ],
                ],
            )
        )

    get_modifications.short_description = "Modifications"

    def get_model(self, instance):
        """
        get the model
        TODO: consider splitting this up to model/pk/value
        """
        return instance.entry.short()

    get_model.short_description = "Model"

    def get_model_link(self, instance):
        """get the model with a link to the entry"""
        return self.model_admin_url(instance.entry)

    get_model_link.short_description = "Model"

    def get_application(self, instance):
        """
        helper to get the application from the child ModelMirror
        :param instance:
        :return:
        """
        return instance.entry.mirror.application

    get_application.short_description = "Application"

    def get_id(self, instance):
        """shorten the id to the first 8 digits"""
        return str(instance.id).split("-")[0]

    get_id.short_description = "UUID"

    def get_user(self, instance):
        """return the user with a link"""
        return self.model_admin_url(instance.user) if instance.user else None

    get_user.short_description = "User"

    list_display = (
        "get_id",
        "updated_at",
        "user",
        "get_application",
        "get_model",
        "get_modifications",
    )

    list_filter = (
        "updated_at",
        ("user", RelatedOnlyFieldListFilter),
        ("entry__mirror__application", RelatedOnlyFieldListFilter),
        ("entry__mirror", RelatedOnlyFieldListFilter),
    )

    date_hierarchy = "updated_at"
    ordering = ("-updated_at",)

    fieldsets = (
        (
            "Information",
            {
                "fields": (
                    "id",
                    "get_user",
                    "updated_at",
                    "get_application",
                    "get_model_link",
                )
            },
        ),
        ("Introspection", {"fields": ("performance", "snapshot")}),
    )
    inlines = [ModelValueModificationInline, ModelRelationshipModificationInline]

    show_change_link = True



""" Test base every unit test uses """

import importlib
import logging.config
from copy import copy, deepcopy

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser
from django.test import TestCase, RequestFactory
from django.urls import path

from automated_logging.helpers import namedtuple2dict
from automated_logging.middleware import AutomatedLoggingMiddleware
from automated_logging.models import ModelEvent, RequestEvent, UnspecifiedEvent
from automated_logging.signals import cached_model_exclusion

User: AbstractUser = get_user_model()
USER_CREDENTIALS = {"username": "example", "password": "example"}


def clear_cache():
    """utility method to clear the cache"""
    if hasattr(AutomatedLoggingMiddleware.thread, "dal"):
        delattr(AutomatedLoggingMiddleware.thread, "dal")

    import automated_logging.decorators

    # noinspection PyProtectedMember
    automated_logging.decorators._exclude_models.clear()
    # noinspection PyProtectedMember
    automated_logging.decorators._include_models.clear()

    cached_model_exclusion.cache_clear()


class BaseTestCase(TestCase):
    def __init__(self, method_name):
        from django.conf import settings

        settings.AUTOMATED_LOGGING_DEV = True

        super().__init__(method_name)

    def request(self, method, view, data=None):
        """
        request a specific view and return the response.

        This is not ideal and super hacky. Backups the actual urlpatterns,
        and then overrides the urlpatterns with a temporary one and then
        inserts the new one again.
        """

        urlconf = importlib.import_module(settings.ROOT_URLCONF)

        backup = copy(urlconf.urlpatterns)
        urlconf.urlpatterns.clear()
        urlconf.urlpatterns.append(path("", view))

        response = self.client.generic(method, "/", data=data)

        urlconf.urlpatterns.clear()
        urlconf.urlpatterns.extend(backup)

        return response

    def setUp(self):
        """setUp the DAL specific test environment"""
        from django.conf import settings
        from automated_logging.settings import default, settings as conf

        self.user = User.objects.create_user(**USER_CREDENTIALS)
        self.user.save()

        self.original_config = deepcopy(settings.AUTOMATED_LOGGING)

        base = namedtuple2dict(default)

        settings.AUTOMATED_LOGGING.clear()
        for key, value in base.items():
            settings.AUTOMATED_LOGGING[key] = deepcopy(value)

        conf.load.cache_clear()

        self.setUpLogging()
        super().setUp()

    def tearDown(self) -> None:
        """tearDown the DAL specific environment"""
        from django.conf import settings
        from automated_logging.settings import settings as conf

        super().tearDown()

        self.tearDownLogging()

        settings.AUTOMATED_LOGGING.clear()
        for key, value in self.original_config.items():
            settings.AUTOMATED_LOGGING[key] = deepcopy(value)

        conf.load.cache_clear()

        clear_cache()

    @staticmethod
    def clear():
        """clear all events"""
        ModelEvent.objects.all().delete()
        RequestEvent.objects.all().delete()
        UnspecifiedEvent.objects.all().delete()

    def tearDownLogging(self):
        """
        replace our own logging config
        with the original to not break any other tests
        that might depend on it.
        """
        from django.conf import settings

        settings.LOGGING = self.logging_backup
        logging.config.dictConfig(settings.LOGGING)

    def setUpLogging(self):
        """sets up logging dict, so that we can actually use our own"""
        from django.conf import settings

        self.logging_backup = deepcopy(settings.LOGGING)
        settings.LOGGING = {
            "version": 1,
            "disable_existing_loggers": False,
            "root": {
                "level": "INFO",
                "handlers": ["console", "db"],
            },
            "formatters": {
                "verbose": {
                    "format": "%(levelname)s %(asctime)s %(module)s "
                    "%(process)d %(thread)d %(message)s"
                },
                "simple": {"format": "%(levelname)s %(message)s"},
                "syslog": {
                    "format": "%(asctime)s %%LOCAL0-%(levelname) %(message)s"
                    # 'format': '%(levelname)s %(message)s'
                },
            },
            "handlers": {
                "console": {
                    "level": "INFO",
                    "class": "logging.StreamHandler",
                    "formatter": "verbose",
                },
                "db": {
                    "level": "INFO",
                    "class": "automated_logging.handlers.DatabaseHandler",
                },
            },
            "loggers": {
                "automated_logging": {
                    "level": "INFO",
                    "handlers": ["console", "db"],
                    "propagate": False,
                },
                "django": {
                    "level": "INFO",
                    "handlers": ["console", "db"],
                    "propagate": False,
                },
            },
        }

        logging.config.dictConfig(settings.LOGGING)

    def bypass_request_restrictions(self):
        """bypass all request default restrictions of DAL"""
        from django.conf import settings
        from automated_logging.settings import settings as conf

        settings.AUTOMATED_LOGGING["request"]["exclude"]["status"] = []
        settings.AUTOMATED_LOGGING["request"]["exclude"]["methods"] = []
        conf.load.cache_clear()

        self.clear()



# Generated by Django 2.0.2 on 2018-02-18 11:01

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("automated_logging", "0011_auto_20180216_1545"),
    ]

    operations = [
        migrations.AddField(
            model_name="request",
            name="method",
            field=models.CharField(default="GET", max_length=64),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="request",
            name="url",
            field=models.CharField(max_length=255),
        ),
    ]



# Generated by Django 2.0.2 on 2018-02-18 11:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("automated_logging", "0012_auto_20180218_1101"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="request",
            name="url",
        ),
        migrations.AddField(
            model_name="request",
            name="uri",
            field=models.URLField(default="/"),
            preserve_default=False,
        ),
    ]



from functools import wraps, partial
from typing import List, NamedTuple, Set, Optional, Type, Dict, Callable, Union


from automated_logging.helpers import (
    Operation,
    get_or_create_thread,
    function2path,
)
from automated_logging.helpers.enums import VerbOperationMap


def _normalize_view_args(methods: List[str]) -> Set[str]:
    if methods is not None:
        methods = {m.upper() for m in methods}

    return methods


# TODO: consider adding status_codes
def exclude_view(func=None, *, methods: Optional[List[str]] = ()):
    """
    Decorator used for ignoring specific views, without adding them
    to the AUTOMATED_LOGGING configuration.

    This is done via the local threading object. This is done via the function
    name and module location.

    :param func: function to be decorated
    :param methods: methods to be ignored (case-insensitive),
                    None => No method will be ignored,
                    [] => All methods will be ignored

    :return: function
    """
    if func is None:
        return partial(exclude_view, methods=methods)

    methods = _normalize_view_args(methods)

    @wraps(func)
    def wrapper(*args, **kwargs):
        """simple wrapper"""
        thread, _ = get_or_create_thread()

        path = function2path(func)
        if (
            path in thread.dal["ignore.views"]
            and thread.dal["ignore.views"][path] is not None
            and methods is not None
        ):
            methods.update(thread.dal["ignore.views"][path])

        thread.dal["ignore.views"][path] = methods

        return func(*args, **kwargs)

    return wrapper


def include_view(func=None, *, methods: List[str] = None):
    """
    Decorator used for including specific views **regardless** if they
    are included in one of the exclusion patterns, this can be selectively done
    via methods. Non matching methods will still go through the exclusion pattern
    matching.

    :param func: function to be decorated
    :param methods: methods to be included (case-insensitive)
                    None => All methods will be explicitly included
                    [] => No method will be explicitly included
    :return: function
    """
    if func is None:
        return partial(include_view, methods=methods)

    methods = _normalize_view_args(methods)

    @wraps(func)
    def wrapper(*args, **kwargs):
        """simple wrapper"""
        thread, _ = get_or_create_thread()

        path = function2path(func)
        if (
            path in thread.dal["include.views"]
            and thread.dal["include.views"][path] is not None
            and methods is not None
        ):
            methods.update(thread.dal["include.views"][path])

        thread.dal["include.views"][path] = methods
        return func(*args, **kwargs)

    return wrapper


def _normalize_model_args(
    operations: List[str], fields: List[str]
) -> [Set[Operation], Set[str]]:
    if operations is not None and not all(
        isinstance(op, Operation) for op in operations
    ):
        operations = {
            VerbOperationMap[o.lower()]
            for o in operations
            if o.lower() in VerbOperationMap.keys()
        }

    if fields is not None:
        fields = set(fields)

    return operations, fields


_include_models = {}
_exclude_models = {}

IgnoreModel = NamedTuple(
    "IgnoreModel", (("operations", Set[Operation]), ("fields", Set[str]))
)


def _register_model(
    registry: Dict[str, Union["IgnoreModel", "IncludeModel"]],
    container: Type[Union["IgnoreModel", "IncludeModel"]],
    decorator: Callable,
    model=None,
    operations: List[str] = None,
    fields: List[str] = None,
):
    if model is None:
        return partial(decorator, operations=operations, fields=fields)

    operations, fields = _normalize_model_args(operations, fields)
    path = function2path(model)

    if (
        path in registry
        and registry[path].operations is not None
        and operations is not None
    ):
        operations.update(registry[path].operations)

    if path in registry and registry[path].fields is not None and fields is not None:
        fields.update(registry[path].fields)

    registry[path] = container(operations, fields)

    # this makes it so that we have a method we can call to re apply dal.
    model.__dal_register__ = lambda: _register_model(
        registry, container, decorator, model, operations, fields
    )
    return model


def exclude_model(
    model=None, *, operations: Optional[List[str]] = (), fields: List[str] = ()
):
    """
    Decorator used for ignoring specific models, without using the
    class or AUTOMATED_LOGGING configuration

    This is done via the local threading object. __module__ and __name__ are used
    to determine the right model.

    :param model: function to be decorated
    :param operations: operations to be ignored can be a list of:
                       modify, create, delete (case-insensitive)
                       [] => All operations will be ignored
                       None => No operation will be ignored
    :param fields: fields to be ignored in not ignored operations
                   [] => All fields will be ignored
                   None => No field will be ignored
    :return: function
    """
    global _exclude_models

    return _register_model(
        _exclude_models, IgnoreModel, exclude_model, model, operations, fields
    )


IncludeModel = NamedTuple(
    "IncludeModel", (("operations", Set[Operation]), ("fields", Set[str]))
)


def include_model(
    model=None, *, operations: List[str] = None, fields: List[str] = None
):
    """
    Decorator used for including specific models, despite potentially being ignored
    by the exclusion preferences set in the configuration.

    :param model: function to be decorated
    :param operations: operations to be ignored can be a list of:
                       modify, create, delete (case-insensitive)
                       [] => No operation will be explicitly included
                       None => All operations will be explicitly included
    :param fields: fields to be explicitly included
                   [] => No fields will be explicitly included
                   None => All fields will be explicitly included.

    :return: function
    """

    global _include_models

    return _register_model(
        _include_models, IncludeModel, include_model, model, operations, fields
    )



"""
Serialization of AUTOMATED_LOGGING_SETTINGS
"""

from collections import namedtuple
from functools import lru_cache
from logging import INFO, NOTSET, CRITICAL
from pprint import pprint

from marshmallow.fields import Boolean, Integer
from marshmallow.validate import OneOf, Range

from automated_logging.helpers.schemas import (
    Set,
    LowerCaseString,
    SearchString,
    MissingNested,
    BaseSchema,
    Search,
    Duration,
)


class RequestExcludeSchema(BaseSchema):
    """
    Configuration schema for request exclusion, that is only used in RequestSchema,
    is used to exclude unknown sources, applications, methods and status codes.
    """

    unknown = Boolean(missing=False)
    applications = Set(SearchString(), missing=set())

    methods = Set(LowerCaseString(), missing={"GET"})
    status = Set(Integer(validate=Range(min=0)), missing={200})


class RequestDataSchema(BaseSchema):
    """
    Configuration schema for request data that is only used in RequestSchema
    and is used to enable data collection, ignore keys that are going to be omitted
    mask keys (their value is going to be replaced with <REDACTED>)
    """

    enabled = Set(
        LowerCaseString(validate=OneOf(["request", "response"])),
        missing=set(),
    )
    query = Boolean(missing=False)

    ignore = Set(LowerCaseString(), missing=set())
    mask = Set(LowerCaseString(), missing={"password"})

    # TODO: add more, change name?
    content_types = Set(
        LowerCaseString(validate=OneOf(["application/json"])),
        missing={"application/json"},
    )


class RequestSchema(BaseSchema):
    """
    Configuration schema for the request module.
    """

    loglevel = Integer(missing=INFO, validate=Range(min=NOTSET, max=CRITICAL))
    exclude = MissingNested(RequestExcludeSchema)

    data = MissingNested(RequestDataSchema)

    ip = Boolean(missing=True)
    # TODO: performance setting?

    log_request_was_not_recorded = Boolean(missing=True)
    max_age = Duration(missing=None)


class ModelExcludeSchema(BaseSchema):
    """
    Configuration schema, that is only used in ModelSchema and is used to
    exclude unknown sources, fields, models and applications.

    fields should be either <field> (every field that matches this name will be excluded),
    or <model>.<field>, or <application>.<model>.<field>

    models should be either <model> (every model regardless of module or application).
    <module> (python module location) or <module>.<model> (python module location)
    """

    unknown = Boolean(missing=False)
    fields = Set(SearchString(), missing=set())
    models = Set(SearchString(), missing=set())
    applications = Set(SearchString(), missing=set())


class ModelSchema(BaseSchema):
    """
    Configuration schema for the model module. mask property indicates
    which fields to specifically replace with <REDACTED>,
    this should be used for fields that are
    sensitive, but shouldn't be completely excluded.
    """

    loglevel = Integer(missing=INFO, validate=Range(min=NOTSET, max=CRITICAL))
    exclude = MissingNested(ModelExcludeSchema)

    # should the log message include all modifications done?
    detailed_message = Boolean(missing=True)

    # if execution_time should be measured of ModelEvent
    performance = Boolean(missing=False)
    snapshot = Boolean(missing=False)

    max_age = Duration(missing=None)


class UnspecifiedExcludeSchema(BaseSchema):
    """
    Configuration schema, that is only used in UnspecifiedSchema and defines
    the configuration settings to allow unknown sources, exclude files and
    specific Django applications
    """

    unknown = Boolean(missing=False)
    files = Set(SearchString(), missing=set())
    applications = Set(SearchString(), missing=set())


class UnspecifiedSchema(BaseSchema):
    """
    Configuration schema for the unspecified module.
    """

    loglevel = Integer(missing=INFO, validate=Range(min=NOTSET, max=CRITICAL))
    exclude = MissingNested(UnspecifiedExcludeSchema)

    max_age = Duration(missing=None)


class GlobalsExcludeSchema(BaseSchema):
    """
    Configuration schema, that is used for every single module.
    There are some packages where it is sensible to have the same
    exclusions.

    Things specified in globals will get appended to the other configurations.
    """

    applications = Set(
        SearchString(),
        missing={
            Search("glob", "session*"),
            Search("plain", "admin"),
            Search("plain", "basehttp"),
            Search("plain", "migrations"),
            Search("plain", "contenttypes"),
        },
    )


class GlobalsSchema(BaseSchema):
    """
    Configuration schema for global, module unspecific configuration details.
    """

    exclude = MissingNested(GlobalsExcludeSchema)


class ConfigSchema(BaseSchema):
    """
    Skeleton configuration schema, that is used to enable/disable modules
    and includes the nested module configurations.
    """

    modules = Set(
        LowerCaseString(validate=OneOf(["request", "model", "unspecified"])),
        missing={"request", "model", "unspecified"},
    )

    request = MissingNested(RequestSchema)
    model = MissingNested(ModelSchema)
    unspecified = MissingNested(UnspecifiedSchema)

    globals = MissingNested(GlobalsSchema)


default: namedtuple = ConfigSchema().load({})


class Settings:
    """
    Settings wrapper,
    with the wrapper we can force lru_cache to be
    cleared on the specific instance
    """

    def __init__(self):
        self.loaded = None
        self.load()

    @lru_cache()
    def load(self):
        """
        loads settings from the schemes provided,
        done via function to utilize LRU cache
        """

        from django.conf import settings as st

        loaded: namedtuple = default

        if hasattr(st, "AUTOMATED_LOGGING"):
            loaded = ConfigSchema().load(st.AUTOMATED_LOGGING)

        # be sure `loaded` has globals as we're working with those,
        # if that is not the case early return.
        if not hasattr(loaded, "globals"):
            return loaded

        # use the binary **or** operator to apply globals to Set() attributes
        values = {}
        for name in loaded._fields:
            field = getattr(loaded, name)
            values[name] = field

            if not isinstance(field, tuple) or name == "globals":
                continue

            values[name] = field | loaded.globals

        self.loaded = loaded._replace(**values)
        return self

    def __getattr__(self, item):
        # self.load() should only trigger when the cache is invalid
        self.load()

        return getattr(self.loaded, item)


@lru_cache()
def load_dev():
    """
    utilize LRU cache and local imports to always
    have an up to date version of the settings

    :return:
    """
    from django.conf import settings as st

    return getattr(st, "AUTOMATED_LOGGING_DEV", False)


if __name__ == "__main__":
    from automated_logging.helpers import namedtuple2dict

    pprint(namedtuple2dict(default))

settings = Settings()
dev = load_dev()



import logging
import threading
from pathlib import Path

from django.contrib import admin
from django.http import JsonResponse

from automated_logging.decorators import (
    include_model,
    exclude_view,
    include_view,
    exclude_model,
)
from automated_logging.helpers import Operation
from automated_logging.middleware import AutomatedLoggingMiddleware
from automated_logging.models import (
    ModelEvent,
    RequestEvent,
    UnspecifiedEvent,
    Application,
)
from automated_logging.signals import (
    cached_model_exclusion,
    model_exclusion,
    request_exclusion,
)
from automated_logging.tests.models import (
    OrdinaryTest,
    OneToOneTest,
    FullClassBasedExclusionTest,
    PartialClassBasedExclusionTest,
    FullDecoratorBasedExclusionTest,
    PartialDecoratorBasedExclusionTest,
    DecoratorOverrideExclusionTest,
    M2MTest,
)
from automated_logging.tests.base import BaseTestCase, USER_CREDENTIALS, clear_cache
from automated_logging.tests.helpers import random_string


class ConfigurationBasedExclusionsTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.client.login(**USER_CREDENTIALS)

    @staticmethod
    def view(request):
        return JsonResponse({})

    def test_globals(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        self.clear()

        settings.AUTOMATED_LOGGING["unspecified"]["exclude"]["applications"] = []
        settings.AUTOMATED_LOGGING["model"]["exclude"]["applications"] = []
        settings.AUTOMATED_LOGGING["request"]["exclude"]["applications"] = []
        settings.AUTOMATED_LOGGING["globals"]["exclude"]["applications"] = [
            "automated*"
        ]

        conf.load.cache_clear()

        OrdinaryTest(random=random_string()).save()
        self.assertEqual(ModelEvent.objects.count(), 0)

        self.request("GET", self.view)
        self.assertEqual(RequestEvent.objects.count(), 0)

        logger = logging.getLogger(__name__)
        logger.info("[TEST]")
        self.assertEqual(UnspecifiedEvent.objects.count(), 0)

    def test_applications(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        self.clear()

        settings.AUTOMATED_LOGGING["globals"]["exclude"]["applications"] = []
        conf.load.cache_clear()

        logger = logging.getLogger(__name__)
        logger.info("[TEST]")
        self.assertEqual(UnspecifiedEvent.objects.count(), 1)
        self.clear()

        settings.AUTOMATED_LOGGING["unspecified"]["exclude"]["applications"] = [
            "automated*"
        ]
        conf.load.cache_clear()

        logger = logging.getLogger(__name__)
        logger.info("[TEST]")
        self.assertEqual(UnspecifiedEvent.objects.count(), 0)

        settings.AUTOMATED_LOGGING["model"]["exclude"]["applications"] = ["automated*"]
        conf.load.cache_clear()

        OrdinaryTest(random=random_string()).save()
        self.assertEqual(ModelEvent.objects.count(), 0)

        settings.AUTOMATED_LOGGING["request"]["exclude"]["applications"] = [
            "automated*"
        ]
        conf.load.cache_clear()

        self.request("GET", self.view)
        self.assertEqual(RequestEvent.objects.count(), 0)

    def test_fields(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        subject = OrdinaryTest()
        subject.save()

        self.clear()
        settings.AUTOMATED_LOGGING["model"]["exclude"]["fields"] = [
            "automated_logging.OrdinaryTest.random"
        ]
        conf.load.cache_clear()

        subject.random = random_string()
        subject.save()
        self.assertEqual(ModelEvent.objects.count(), 0)

        settings.AUTOMATED_LOGGING["model"]["exclude"]["fields"] = [
            "OrdinaryTest.random"
        ]
        conf.load.cache_clear()

        subject.random = random_string()
        subject.save()
        self.assertEqual(ModelEvent.objects.count(), 0)

        settings.AUTOMATED_LOGGING["model"]["exclude"]["fields"] = ["random"]
        conf.load.cache_clear()
        subject.random = random_string()
        subject.save()
        self.assertEqual(ModelEvent.objects.count(), 0)

        subject.random = random_string()
        subject.random2 = random_string()
        subject.save()
        self.assertEqual(ModelEvent.objects.count(), 1)
        self.assertEqual(ModelEvent.objects.all()[0].modifications.count(), 1)

    def test_models(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        self.clear()

        settings.AUTOMATED_LOGGING["model"]["exclude"]["models"] = [
            "automated_logging.tests.models.OrdinaryTest"
        ]
        conf.load.cache_clear()

        OrdinaryTest(random=random_string()).save()
        self.assertEqual(ModelEvent.objects.count(), 0)

        settings.AUTOMATED_LOGGING["model"]["exclude"]["models"] = [
            "automated_logging.OrdinaryTest"
        ]
        conf.load.cache_clear()

        OrdinaryTest(random=random_string()).save()
        self.assertEqual(ModelEvent.objects.count(), 0)

        settings.AUTOMATED_LOGGING["model"]["exclude"]["models"] = ["OrdinaryTest"]
        conf.load.cache_clear()

        OrdinaryTest(random=random_string()).save()
        self.assertEqual(ModelEvent.objects.count(), 0)

        OneToOneTest().save()
        self.assertEqual(ModelEvent.objects.count(), 1)
        self.assertEqual(ModelEvent.objects.all()[0].modifications.count(), 1)

    @staticmethod
    def redirect_view(request):
        return JsonResponse({}, status=301)

    def test_status(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        settings.AUTOMATED_LOGGING["request"]["exclude"]["methods"] = []
        settings.AUTOMATED_LOGGING["request"]["exclude"]["status"] = [200]
        conf.load.cache_clear()

        self.clear()

        self.request("GET", self.view)
        self.assertEqual(RequestEvent.objects.count(), 0)

        self.request("GET", self.redirect_view)
        self.assertEqual(RequestEvent.objects.count(), 1)

    def test_method(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        settings.AUTOMATED_LOGGING["request"]["exclude"]["methods"] = ["GET"]
        settings.AUTOMATED_LOGGING["request"]["exclude"]["status"] = []
        conf.load.cache_clear()

        self.clear()

        self.request("GET", self.view)
        self.assertEqual(RequestEvent.objects.count(), 0)

        self.request("POST", self.view)
        self.assertEqual(RequestEvent.objects.count(), 1)

    def test_files(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        path = Path(__file__).absolute()
        project = Path(__file__).parent.parent.parent
        relative = path.relative_to(project)

        logger = logging.getLogger(__name__)
        self.clear()

        # absolute path
        settings.AUTOMATED_LOGGING["unspecified"]["exclude"]["files"] = [
            path.as_posix()
        ]
        conf.load.cache_clear()

        logger.info(random_string())
        self.assertEqual(UnspecifiedEvent.objects.count(), 0)

        # relative path
        settings.AUTOMATED_LOGGING["unspecified"]["exclude"]["files"] = [
            relative.as_posix()
        ]
        conf.load.cache_clear()

        logger.info(random_string())
        self.assertEqual(UnspecifiedEvent.objects.count(), 0)

        # file name
        settings.AUTOMATED_LOGGING["unspecified"]["exclude"]["files"] = [relative.name]
        conf.load.cache_clear()

        logger.info(random_string())
        self.assertEqual(UnspecifiedEvent.objects.count(), 0)

        # single directory name
        settings.AUTOMATED_LOGGING["unspecified"]["exclude"]["files"] = [
            "automated_logging"
        ]
        conf.load.cache_clear()

        logger.info(random_string())
        self.assertEqual(UnspecifiedEvent.objects.count(), 0)

        # absolute directory
        settings.AUTOMATED_LOGGING["unspecified"]["exclude"]["files"] = [
            path.parent.as_posix()
        ]
        conf.load.cache_clear()

        logger.info(random_string())
        self.assertEqual(UnspecifiedEvent.objects.count(), 0)

        # file not excluded
        settings.AUTOMATED_LOGGING["unspecified"]["exclude"]["files"] = ["dal"]
        conf.load.cache_clear()

        logger.info(random_string())
        self.assertEqual(UnspecifiedEvent.objects.count(), 1)

    def test_unknown(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        logger = logging.getLogger(__name__)

        default_factory = logging.getLogRecordFactory()

        def factory(*args, **kwargs):
            """
            force setting the pathname and module
            wrong so that we can pretend to exclude unknowns
            """

            record = default_factory(*args, **kwargs)

            record.pathname = "/example.py"
            record.module = "default"
            return record

        self.clear()
        logging.setLogRecordFactory(factory=factory)

        settings.AUTOMATED_LOGGING["unspecified"]["exclude"]["unknown"] = True
        conf.load.cache_clear()

        logger.info(random_string())
        self.assertEqual(UnspecifiedEvent.objects.count(), 0)

        settings.AUTOMATED_LOGGING["unspecified"]["exclude"]["unknown"] = False
        conf.load.cache_clear()

        logger.info(random_string())
        self.assertEqual(UnspecifiedEvent.objects.count(), 1)

        logging.setLogRecordFactory(default_factory)

    def test_search_types(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        settings.AUTOMATED_LOGGING["globals"]["exclude"]["applications"] = []
        settings.AUTOMATED_LOGGING["model"]["exclude"]["applications"] = [
            "pl:automated_logging"
        ]
        conf.load.cache_clear()
        cached_model_exclusion.cache_clear()

        self.clear()
        OrdinaryTest().save()
        self.assertEqual(ModelEvent.objects.count(), 0)

        settings.AUTOMATED_LOGGING["model"]["exclude"]["applications"] = [
            "gl:automated_*"
        ]
        conf.load.cache_clear()
        cached_model_exclusion.cache_clear()

        OrdinaryTest().save()
        self.assertEqual(ModelEvent.objects.count(), 0)

        settings.AUTOMATED_LOGGING["request"]["exclude"]["applications"] = [
            "re:automated.*"
        ]
        self.bypass_request_restrictions()
        conf.load.cache_clear()
        cached_model_exclusion.cache_clear()

        self.request("GET", self.view)
        self.assertEqual(RequestEvent.objects.count(), 0)

    def test_module(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        settings.AUTOMATED_LOGGING["globals"]["exclude"]["applications"] = []
        settings.AUTOMATED_LOGGING["model"]["exclude"]["models"] = [
            "automated_logging.tests.models"
        ]
        conf.load.cache_clear()
        cached_model_exclusion.cache_clear()
        self.clear()

        OrdinaryTest().save()
        self.assertEqual(ModelEvent.objects.count(), 0)

    def test_mock_sender(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        class MockModel:
            __module__ = "[TEST]"
            __name__ = "MockModel"

        class MockMeta:
            app_label = None

        settings.AUTOMATED_LOGGING["model"]["exclude"]["unknown"] = True
        conf.load.cache_clear()

        self.assertTrue(model_exclusion(MockModel, MockMeta, Operation.CREATE))

        settings.AUTOMATED_LOGGING["request"]["exclude"]["unknown"] = True
        conf.load.cache_clear()

        self.assertTrue(
            request_exclusion(RequestEvent(application=Application(name=None)))
        )

    def test_m2m_exclusion(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        settings.AUTOMATED_LOGGING["model"]["exclude"]["applications"] = [
            "gl:automated*"
        ]
        conf.load.cache_clear()
        cached_model_exclusion.cache_clear()

        instance = M2MTest()
        instance.save()

        children = [OrdinaryTest()]
        [c.save() for c in children]

        self.clear()

        instance.relationship.add(*children)
        instance.save()

        self.assertEqual(ModelEvent.objects.count(), 0)


class ClassBasedExclusionsTestCase(BaseTestCase):
    def test_complete(self):
        self.clear()

        FullClassBasedExclusionTest().save()
        self.assertEqual(ModelEvent.objects.count(), 0)

    def test_partial(self):
        self.clear()

        subject = PartialClassBasedExclusionTest(random=random_string())
        subject.save()

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertEqual(event.modifications.count(), 1)
        self.assertEqual(event.modifications.all()[0].field.name, "id")

        self.clear()

        subject.delete()
        self.assertEqual(ModelEvent.objects.count(), 0)


class DecoratorBasedExclusionsTestCase(BaseTestCase):
    def test_exclude_model(self):
        self.clear()

        # manual here as the cache clear, clears the first appliance
        FullDecoratorBasedExclusionTest.__dal_register__()
        subject = FullDecoratorBasedExclusionTest()
        subject.save()

        self.assertEqual(ModelEvent.objects.count(), 0)

        subject = PartialDecoratorBasedExclusionTest.__dal_register__()(
            random=random_string()
        )
        subject.save()

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertEqual(event.modifications.count(), 1)

        self.clear()
        subject.delete()

        self.assertEqual(ModelEvent.objects.count(), 0)

    def test_include_model(self):
        self.clear()

        subject = DecoratorOverrideExclusionTest.__dal_register__()(
            random=random_string()
        )
        subject.save()

        self.assertEqual(ModelEvent.objects.count(), 1)

        self.clear()

        # test if overriding works
        include_model(FullClassBasedExclusionTest, operations=["delete"])()

        subject = FullClassBasedExclusionTest(random=random_string())
        subject.save()

        self.assertEqual(ModelEvent.objects.count(), 0)

        subject.delete()

        self.assertEqual(ModelEvent.objects.count(), 1)

        # just to make sure we clean up
        clear_cache()

    @staticmethod
    @exclude_view
    def complete_exclusion(request):
        return JsonResponse({})

    @staticmethod
    @exclude_view(methods=["GET"])
    def partial_exclusion(request):
        return JsonResponse({})

    def test_exclude_view(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        settings.AUTOMATED_LOGGING["request"]["exclude"]["methods"] = []
        settings.AUTOMATED_LOGGING["request"]["exclude"]["status"] = []
        conf.load.cache_clear()
        self.clear()

        self.request("GET", self.complete_exclusion)
        self.assertEqual(RequestEvent.objects.count(), 0)

        self.request("GET", self.partial_exclusion)
        self.assertEqual(RequestEvent.objects.count(), 0)
        self.request("POST", self.partial_exclusion)
        self.assertEqual(RequestEvent.objects.count(), 1)

    @staticmethod
    @include_view
    def complete_inclusion(request):
        return JsonResponse({})

    @staticmethod
    @include_view(methods=["POST"])
    def partial_inclusion(request):
        return JsonResponse({})

    def test_include_view(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        self.clear()
        # settings default to ignoring 200/GET, include_model should still record

        self.request("GET", self.complete_inclusion)
        self.assertEqual(RequestEvent.objects.count(), 1)
        self.clear()

        settings.AUTOMATED_LOGGING["request"]["exclude"]["status"] = []
        settings.AUTOMATED_LOGGING["request"]["exclude"]["methods"] = ["GET", "POST"]
        conf.load.cache_clear()

        self.request("GET", self.partial_inclusion)
        self.assertEqual(RequestEvent.objects.count(), 0)
        self.request("POST", self.partial_inclusion)
        self.assertEqual(RequestEvent.objects.count(), 1)
        self.clear()

        settings.AUTOMATED_LOGGING["request"]["exclude"]["methods"] = []
        conf.load.cache_clear()

        # test if include_view has higher priority than exclude_view
        view = include_view(self.complete_exclusion, methods=["GET"])
        self.request("GET", view)
        self.assertEqual(RequestEvent.objects.count(), 1)
        self.request("POST", view)
        self.assertEqual(RequestEvent.objects.count(), 1)

        clear_cache()

    def test_partial_decorator(self):
        self.clear()
        Model = include_model(operations=["create"])(OrdinaryTest)
        Model().save()
        self.assertEqual(ModelEvent.objects.count(), 1)

    def test_layering(self):
        Model = include_model(operations=["modify"], fields=["random2"])(
            include_model(operations=["create"], fields=["random2"])(OrdinaryTest)
        )

        self.clear()
        subject = Model()
        subject.save()

        subject.random = random_string()
        subject.save()

        self.assertEqual(ModelEvent.objects.count(), 2)

        clear_cache()

        Model = exclude_model(operations=["modify"], fields=["random2"])(
            exclude_model(operations=["create"], fields=["random2"])(OrdinaryTest)
        )

        self.clear()

        subject = Model()
        subject.save()

        subject.random = random_string()
        subject.save()

        self.assertEqual(ModelEvent.objects.count(), 0)

    def test_model_fields(self):
        Model = include_model(operations=[], fields=["random"])(
            exclude_model(operations=None, fields=["random2"])(OrdinaryTest)
        )

        subject = Model()
        subject.save()

        self.clear()

        subject.random = random_string()
        subject.save()

        self.assertEqual(ModelEvent.objects.count(), 1)

        subject.random2 = random_string()
        subject.save()

        self.assertEqual(ModelEvent.objects.count(), 1)

    def test_model_admin(self):
        FullDecoratorBasedExclusionTest.__dal_register__()

        admin.site.register(FullDecoratorBasedExclusionTest)
